'use strict';

const pg = require('pg');
const path = require('path');
const fs = require('fs')

const types = require('pg').types;
const moment = require("moment")
var parseFn = function (val) {
    return val === null ? null : moment(val).format('YYYY-MM-DD HH:mm:ss.SSSSSS')
}

//types.setTypeParser(types.builtins.TIMESTAMPTZ, parseFn)
//types.setTypeParser(types.builtins.TIMESTAMP, parseFn)
//types.setTypeParser(types.builtins.DATE, parseFn)
const utils = require('./zmhutils')

// 数据库配置
const config = {
    user: "postgres",
    database: "postgres",
    password: "postgres",
    host: '127.0.0.1',
    port: 5432,
    max: 10, // 连
    //types,
    // 接池最大连接数
    idleTimeoutMillis: 30000, // 连接最大空闲时间 3s
};
//const logger = require('../../utils/logger')

class Db {
    constructor(options, filename) {
        this.op = Object.assign({}, config, options)
        let f = path.parse(filename)
        this.dbname = f.name
        this.dbhome=f.dir
        this.dbdir = path.join(f.dir, f.name)
        if (!fs.existsSync(this.dbdir)) { fs.mkdirSync(this.dbdir) }

        this.pool = new pg.Pool(this.op)
    }

    async execute(sql, param) {
        // console.log(sql)
        let conn;
        try {
            conn = await this.pool.connect();
            let result = await conn.query(sql, param);
            return { code: 1, result };
        }
        catch (err) {
            console.log(sql)
            console.log(err + '');
            return { code: -3, err: err + '' }
        } finally {
            conn.release();
        };
    };

    async execute1({ sql, param }) {
        return this.execute(sql, param)
    };
    async executes(sql) {
        // console.log(sql)
        let conn;
        try {
            conn = await this.pool.connect();
            let r = await conn.query(sql);
            //console.log(r)
            return { code: 1, data: "ok" }
        }
        catch (err) {
            //console.log(err);
            return { code: -3, err: err + '' }
        } finally {
            conn.release();
        };
    };

    async dtproc(p) {
        let t = await this.getschemafile(p)
        //返回 日期类型字段列表
        //console.log(t)
        for (let i in t.cols) {
            if (t.cols[i].CTYPE == 'timestamp' || t.cols[i].CTYPE == 'timestamptz') {
                p.data.map(j => {
                    if (j[i]) {
                        j[i] = moment(new Date(j[i])).format('YYYY-MM-DD HH:mm:ss.SSS')
                        //console.log(j)
                        return j
                    }
                })
            }
            if (t.cols[i].CTYPE == 'date') {
                p.data.map(j => {
                    if (j[i]) {
                        j[i] = moment(new Date(j[i])).format('YYYY-MM-DD')
                        //console.log(j)
                        return j
                    }
                })
            }

        }
        // console.log(p)
        //let dts = t.colsa.filter(i => { if (i.CTYPE == 'DATE') { return i } }).map(i => i.CNAME)
        // console.log(dts)
        return p
    }

    getexpression(p) {
        let s = ''
        switch (p.operator) {
            case 'startwith':
                if (p.value != null) {
                    s = `("${p.field}"::text like '${p.value}%')`
                } else {
                    s = `(true)`
                }
                break;
            case 'endswith':
                if (p.value != null) {
                    s = `("${p.field}"::text like '%${p.value}')`
                } else {
                    s = `(true)`
                }
                break;
            case 'contains':
                if (p.value != null) {
                    s = `("${p.field}"::text like '%${p.value}%')`
                } else {
                    s = `(true)`
                }
                break;
            case 'isnull':
            case 'isempty':
                s = `("${p.field}" is null)`
                break;
            case 'isnotnull':
            case 'isnotempty':
                s = `("${p.field}" is not null)`
                break;
            case 'eq':
                if (p.value != null) {
                    s = `("${p.field}" = '${p.value}')`
                } else {
                    s = `("${p.field}" is null)`
                }
                break;
            case 'neq':
                if (p.value != null) {
                    s = `("${p.field}" <> '${p.value}')`
                } else {
                    s = `("${p.field}" is not null)`
                }
                break;
            case 'lt':
                if (p.value != null) {
                    s = `("${p.field}" < '${p.value}')`
                } else {
                    s = `(true)`
                }
                break;
            case 'lte':
                if (p.value != null) {
                    s = `("${p.field}" <= '${p.value}')`
                } else {
                    s = `(true)`
                }
                break;
            case 'gt':
                if (p.value != null) {
                    s = `("${p.field}" > '${p.value}')`
                } else {
                    s = `(true)`
                }
                break;
            case 'gte':
                if (p.value != null) {
                    s = `("${p.field}" >= '${p.value}')`
                } else {
                    s = `(true)`
                }
                break;
        }
        return s
    }

    getwhere(f) {
        let a = []
        for (let i = 0; i < f.filters.length; i++) {
            let f1 = f.filters[i]
            if (f1.filters) {
                a.push(this.getwhere(f1))
            }
            else {
                //是一个表达式
                a.push(' ' + this.getexpression(f1) + ' ')
                //console.log(f1)
            }
        }
        let s = ' (' + a.join(f.logic) + ') '
        return s
    }

    typecast(vname, value, ctype) {
        let r = { value: `'${value}'`, vname: `"${vname}"` }
        if (value == null) {
            r.value = `null`
        }
        switch (ctype) {
            case 'VARCHAR2':
            case 'NVARCHAR2':
                break;
            case 'DATE':
                break;
            case 'json':
                r.value = `'${JSON.stringify(value)}'`;
                break;
        }
        return r
    }

    async getcolumnsbyschema(p) {
        let schema = p.schema
        if (!schema) { schema = 'public' }
        let sql = `
        select t1.cname "CNAME", coalesce(t1.descript,t1.cname)  "REMARK" ,t2.is_nullable "CNULL" ,t1.data_type "DATATYPE",t2.udt_name "CTYPE", coalesce(t2.character_maximum_length,numeric_precision) "CLEN" from (select a.attnum ,
            (select description from pg_catalog.pg_description where objoid=a.attrelid and objsubid=a.attnum) as descript ,
            a.attname cname,
            pg_catalog.format_type(a.atttypid,a.atttypmod) as data_type,
            b.sname,
            b.tname
           from 
             pg_catalog.pg_attribute a ,
             (select a1.oid,a1.relname tname,a2.nspname sname from pg_class a1, pg_namespace a2  where a1.relname='${p.tname}' and a1.relnamespace = a2.oid and 
                  a2.nspname = '${schema}') as b
            where a.attrelid=b.oid and a.attnum>=0 and not a.attisdropped 
            ) t1,
            information_schema.columns t2
              where t1.sname =t2.table_schema and t1.tname = t2.table_name and t2.column_name = t1.cname
       `
        //console.log(sql)
        let r = await this.execute(sql)
        return r
    }

    async createlogtable(p) {
        let schema = p.schema
        if (!schema) { schema = 'public' }
        let sql = `drop type ${p};create type p.tname as 
        
        `

    }

    getclass(o) {
        if (o === null) return "Null";
        if (o === undefined) return "undefined";
        return Object.prototype.toString.call(o).slice(8, -1);
    }

    async executesqls(sqls) {
        let r = []
        let code = 1
        for (let i in sqls) {
            let sql = sqls[i]
            let r0 = await this.execute(sql)
            if (r0.code == 1) {
                r.push({ code: 1, sql, result: r0.result })
            } else {
                code = 2;
                r.push({ ...r0, sql })
            }
        }
        return { code, data: r }
    }

    async createformdatatable(p = {}) {
        let p0 = { schema: 'public' }
        p = Object.assign(p, p0)
        let sqls = [
            `drop TABLE "${p.schema}"."sys_formdata"`,
            `drop SEQUENCE "${p.schema}"."sys_formdata_id_seq"`,
            `create SEQUENCE "${p.schema}"."sys_formdata_id_seq"`,
            `CREATE TABLE "${p.schema}"."sys_formdata" (
              "__id" int8 NOT NULL DEFAULT nextval('sys_formdata_id_seq'::regclass),
              "tname" varchar(120),
              "data" jsonb,
              "remark" varchar(200),
              CONSTRAINT "sys_formdata_pkey" PRIMARY KEY ("__id")
            )`
        ]
        return await this.executesqls(sqls)
    }

    async createtypewithschema(p = {}) {
        let p0 = { schema: 'public' }
        p = Object.assign(p, p0)
        let cols0 = p.cols.map(i => {
            let r = { name: i.tname }
            switch (i.type) {
                case 'Number':
                    r.type = "decimal"
                    break
                case 'Date':
                    r.type = "timestamp"
                    break
                case 'Object':
                    r.type = "json"
                    break
                default:
                    r.type = "text"
            }
            return r
        })
        let c = cols0.map(i => { return `"${i.name}" ${i.type}` }).join(',')
        let sqls = [
            `drop type "${p.schema}"."form_${p.tname}"`,
            `create type "${p.schema}"."form_${p.tname}" as(${c})`,
        ]
        return await this.executesqls(sqls)
    }

    async createtype(p) {
        let schema = p.schema
        let list = []
        if (!schema) { schema = 'public' }
        if (this.getclass(p.json) == 'Object') {
            for (let key in p.json) {
                switch (this.getclass(p.json[key])) {
                    case 'Number':
                        list.push(`"${key}" decimal`)
                        break
                    case 'Date':
                        list.push(`"${key}" timestamp`)
                        break
                    default:
                        list.push(`"${key}" text`)
                }
            }
            let sql = `drop type if exists ${p.tname} ;
            create type ${p.tname} as (${list})
           `
            let r = await this.execute(sql)
            return r
        }
    }

    async createtable(p) {
        let schema = p.schema
        let list = []
        if (!schema) { schema = 'public' }
        if (this.getclass(p.json) == 'Object') {
            for (let key in p.json) {
                switch (this.getclass(p.json[key])) {
                    case 'Number':
                        list.push(`"${key}" decimal`)
                        break
                    case 'Date':
                        list.push(`"${key}" timestamp`)
                        break
                    case 'Object':
                    case 'Array':
                        list.push(`"${key}" json`)
                        break
                    default:
                        list.push(`"${key}" text`)
                }
            }
            let sql = `drop table if exists ${p.tname} ;
            create table ${p.tname} (${list})
           `
            console.log(sql)
            let r = await this.execute(sql)
            return r
        }
    }



    async getpkcols(p) {
        let schema = p.schema
        if (!schema) { schema = 'public' }
        let sql = `
       select pg_attribute.attname as colname
       from pg_constraint  
             inner join (select a1.oid,a1.relname,a2.nspname from pg_class a1,pg_namespace a2 
                    where a1.relnamespace =a2.oid and a1.relname = '${p.tname}' and a2.nspname ='${schema}' ) t
           on pg_constraint.conrelid = t.oid 
         inner join pg_attribute 
                 on pg_attribute.attrelid = t.oid and pg_constraint.conkey @>array[pg_attribute.attnum ]
         inner join pg_type on pg_type.oid = pg_attribute.atttypid
       where pg_constraint.contype='p'       
       `
        let r = await this.execute(sql)
        if (r && r.result.rows) {
            let pkcols = r.result.rows.map(i => { return i.colname })
            r = { pkcols }
        }
        return r
    }

    async gettabsbyschema(p) {
        let schema = p.schema
        if (!schema) { schema = 'public' }
        let sql = `
        select tablename tname from pg_tables where schemaname ='${schema}' 
       `
        let r = await this.execute(sql)
        //console.log(r)
        if (r.code == 1) {
            return r.rows.map(i => { return i.tname })
        }
        return []
    }

    //从数据库中获取定义默认值值得字段列表
    //参数{tname 表名称,schema 模式名称}
    async getdefaultcols(p) {
        let schema = p.schema
        if (!schema) { schema = 'public' }
        let sql = `
        select a.attname cname,'' define from pg_attribute a,
        pg_attrdef c,
        (select a.oid,relname tname,nspname sname from pg_class a, pg_namespace b 
           where b.oid =a.relnamespace and nspname= '${schema}' and a.relname='${p.tname}'
        ) b 
        where a.attrelid = b.oid and a.attnum>=0 and not a.attisdropped  
                and c.adrelid = b.oid and c.adnum = a.attnum
       `
        //console.log(sql)
        let r = await this.execute(sql)
        //console.log(r)0
        return r
    }


    //{schema:'public',tname:'sys_users',func:'_tr_func_dbop_log',cols:[]}
    async createdboplogtr(p) {
        let cols = ""
        if (p.cols && p.cols.length > 0) { cols = ' OF "' + p.cols.join('","') + '"' }
        let p0 = { schema: 'public', func: '_tr_func_dbop_log' }
        p = Object.assign(p, p0)

        let sql
        sql = `drop trigger  "trdblog01_${p.tname}" on "${p.schema}"."${p.tname}"`
        await this.execute(sql);
        sql = `
    CREATE TRIGGER "trdblog01_${p.tname}" AFTER INSERT OR UPDATE ${cols} OR DELETE ON "${p.schema}"."${p.tname}"
    FOR EACH ROW
    EXECUTE PROCEDURE "${p.schema}"."${p.func}"();
    `;
        return await this.execute(sql);
    }

    // {p:{schema}}
    async createdboplog(p) {
        let schema = p.schema
        if (!schema) { schema = 'public' }
        let sql
        sql = `drop table "${schema}"."_tr_dbop_log"`
        await this.execute(sql)

        sql = `drop SEQUENCE "${schema}"."_tr_dbop_log_id_seq"`
        await this.execute(sql)
        sql = `CREATE SEQUENCE "${schema}"."_tr_dbop_log_id_seq"  INCREMENT 1 MINVALUE  1 MAXVALUE 9223372036854775807 START 1 CACHE 1`;
        await this.execute(sql)

        sql = `
        CREATE TABLE "${schema}"."_tr_dbop_log" (
            "id" int8 NOT NULL DEFAULT nextval('"${schema}"."_tr_dbop_log_id_seq"'::regclass),
            "tname" varchar(120) COLLATE "pg_catalog"."default",
            "dt0" timestamp(6) DEFAULT now(),
            "otype" varchar(20) COLLATE "pg_catalog"."default",
            "newdata" json,
            "olddata" json,
            "status0" int2 DEFAULT 0,
            "status1" int2 DEFAULT 0,
            "status2" int2 DEFAULT 0,
            CONSTRAINT "_tr_dbop_log_pkey" PRIMARY KEY ("id")
          )
          ;`
        await this.execute(sql)

        let sql01 =

            `CREATE OR REPLACE FUNCTION "${schema}"."_tr_func_dbop_log"()
        RETURNS "pg_catalog"."trigger" AS $BODY$
        BEGIN

        IF TG_OP='INSERT' THEN
            insert into "${schema}"."_tr_dbop_log"(tname,otype,newdata) VALUES(TG_TABLE_NAME,TG_OP,row_to_json(new));
                return NEW;
        END IF;
        IF TG_OP='UPDATE' THEN
            insert into "${schema}"."_tr_dbop_log"(tname,otype,newdata,olddata) VALUES(TG_TABLE_NAME,TG_OP,row_to_json(new),row_to_json(old));
                return NEW;
        END IF;
        IF TG_OP='DELETE' THEN
            insert into "${schema}"."_tr_dbop_log"(tname,otype,olddata) VALUES(TG_TABLE_NAME,TG_OP,row_to_json(old));
                return old;
        END IF;
        END ;
        $BODY$
        LANGUAGE plpgsql VOLATILE
        COST 100`
        let r = await this.execute(sql01)
        return r
    }



    getPath(p, type,dbname) {
        //console.log(cnf.rootDir, '/sys/config/db/', this.dbname, type, tname)
        let filename
        let dir = path.join(this.dbdir, type)
        if (dbname) dir = path.join(this.dbhome,dbname,type)        
        //console.log(dir)
        try {
            if (dbname) {
                let dir0 =path.join(this.dbhome,dbname)
                if  (!fs.existsSync(dir0)) {
                    fs.mkdirSync(dir0)
                }
            }
            if (!fs.existsSync(dir)) {
                fs.mkdirSync(dir)
            }
        } catch (e) {
            console.log(e)
            return null
        }
        if (!p.schema) { filename = p.tname } else { filename = p.schema + '.' + p.tname }
        return path.join(dir, filename)
    }

    async getschemafile(p) {
        if (!p.schema) { p.schema = 'public' }
        try {
            let file = this.getPath(p, '/tables')
            if (fs.existsSync(file + '.js')) {
                return require(file)
            } else {
                await this.makeTableSchema(p)
                return require(file)
            }
        } catch (err) {
            console.log(err)
            return null

        }
    }



    async getsqlfile(p) {
        try {
            let file = this.getPath(p, '/sqls')
            if (fs.existsSync(file + '.js')) {
                return require(file)
            } else {
                return null
            }
        } catch (err) {
            console.log(err)
            return null
        }
    }

    async listsqlfiles(p) {
        //console.log(p)
        try {
            let p1 = path.join(this.dbdir, 'sqls')
            let r = utils.getFileList(p1)
            r.map(i => {
                try {
                    let i0 = require(i.fullpath)
                    i.title = i0.title,
                        i.template = i0.template
                    i.remark = i0.remark
                    i.param = i0.param
                    i.fullpath = ''

                    return i
                } catch (e) {
                    console.log(e)
                }
            })
            return r
        } catch (err) {
            console.log(err)
            return null
        }
    }



    async savesqlfile(p) {
        try {
            //console.log(p)
            let p0 = { title: '', remark: '' }
            p = Object.assign(p, p0)
            let file = this.getPath(p, '/sqls',p.dbname)
            let tmp = `
const art =require('art-template');
const utils =require('zmhutils');

let obj ={}
obj.template = \`${p.template}\`
obj.template0 = utils.tempProc(obj.template,__filename)
obj.param =${JSON.stringify(p.param, null, "\t")}
obj.render =art.compile(obj.template0)
obj.title = '${p.title}',
obj.remark ='${p.remark}'
module.exports = obj
`
            fs.writeFileSync(file + '.js', tmp)
            utils.require2(file)
            //console.log(a)
            return { code: 1, message: 'write ok' }
        } catch (err) {
            console.log(err)
            return null
        }
    }

    async saveSqlById({id}) {
        try {
           let  r =await this.getTableData({tname:'sys_sql',data:[{id}]})

           if (r.code ==1 && r.data.length > 0) {
            let p= r.data[0]
            p.tname = p.name
            let file = this.getPath(p, '/sqls',p.dbname)
            let tmp = `
const art =require('art-template');
const utils =require('zmhutils');
let obj ={}
obj.template = \`${p.sql}\`
obj.template0 = utils.tempProc(obj.template,__filename)
obj.param =${JSON.stringify(p.params, null, "\t")}
obj.render =art.compile(obj.template0)
obj.title = '${p.name}',
obj.remark ='${p.remark}'
module.exports = obj
`
            fs.writeFileSync(file + '.js', tmp)
            utils.require2(file)
            //console.log(a)
            return { code: 1, message: 'write ok' }

           }
            //console.log(p)
        } catch (err) {
            console.log(err)
            return null
        }
    }

    async delSqlById({id}) {
        try {
           let  r =await this.getTableData({tname:'sys_sql',data:[{id}]})

           if (r.code ==1 && r.data.length > 0) {
            let p= r.data[0]
            p.tname = p.name
            let file = this.getPath(p, '/sqls')
            fs.rmSync(file + '.js')
            delete require.cache[require.resolve(file)];
            return { code: 1, message: 'delete ok' }
           }
            //console.log(p)
        } catch (err) {
            console.log(err)
            return null
        }
    }

    async sqlRun({tname,params}) {
        try {
            let file = this.getPath({tname}, '/sqls')
            let {render} = require(file)
            let sql =render(params) 
            let r1 = await this.execute(sql)
            r1.sql =sql
            if (r1.code == 1 && r1.result.rowCount > 0) { return { code: 1, data: r1.result.rows } }
            return r1
        } catch (err) {
            console.log(err)
            return {code:-3,err}
        }
    }

    async sqlTest({tname,params}) {
        try {
            let file = this.getPath({tname}, '/sqls')
            let {render} = require(file)
            let sql =render(params) 
            let r1 = await this.execute(sql)
            r1.sql =sql
            if (r1.code == 1 && r1.result.rowCount > 0) { return { sql:r1.sql,code: 1, data: r1.result.rows } }
            return r1
        } catch (err) {
            console.log(err)
            return {code:-3,err}
        }
    }
    
    async  sqlInsert({ schema = 'public', tname, data, keycase, user, log }){
        let r =await this.tableInsert({ schema, tname, data, keycase, user, log })
        if (r.code == 1 && r.data && r.data.length > 0){
            r.data.forEach(i=> this.saveSqlById({id:i.id}))
        }
        return r 
    }

    async  sqlDelete({ schema = 'public', tname, data, keycase, user, log }){
        let r =await this.tableDelete({ schema, tname, data, keycase, user, log })
        if (r.code == 1 && r.data && r.data.length > 0){
            r.data.forEach(i=> this.delSqlById({id:i.id}))
        }
        return r 

    }
    async  sqlUpdate({ schema = 'public', tname, data, keycase, user, log }){
        let r =await this.tableUpdate({ schema, tname, data, keycase, user, log })
        if (r.code == 1 && r.data && r.data.length > 0){
            r.data.forEach(i=>{  this.saveSqlById({id:i.id})   })
        }
        return r 
    }

    async getnextseqno(p) {
        let sql = `select nextval('${p.seqname}') as seqno`
        let r = await this.execute(sql)
        if (r.code > 0) {
            return r.rows[0].seqno;
        } else {
            return -1
        }
    }
    async preprocdata(data, method, p) {
        let r = { cols: [], data: {}, data0: [], pkey: [], updates: [], cols1: [], pkcols: [] }
        try {
            let t = await this.getschemafile(p)
            //console.log(t.pkcols)
            r.t = t
            if (method == 'INSERT') {
                if (t.beforeInsert) {
                    //console.log(t.beforeInsert.toString())
                    data = await t.beforeInsert(this, data)
                }
            }
            for (let key in data) {
                let col = t.cols[key]
                if (col) {
                    r.data[key] = data[key]
                    let r2 = this.typecast(key, data[key], col.CTYPE)
                    // console.log(r2)
                    let colname = key
                    r.data0.push(r2.value)

                    if (t.pkcols.indexOf(colname) >= 0) {
                        r.pkey.push(r2.vname + '=' + r2.value)
                        r.pkcols.push(r2.vname)
                    } else {
                        r.updates.push(r2.vname + '=' + r2.value)
                    }
                    r.cols.push(colname)
                    r.cols1.push(r2.vname)
                }
            }
            //console.log(r)
            return r;

        } catch (err) {
            console.log(err)
            return { errcode: -1, err }
        }
    }

    //参数实例：{tname:'t1',schema:'public'}
    async makeTableSchema(p) {
        // console.log(p)
        try {
            let schema = {}
            if (!p.schema) { p.schema = 'public' }
            let r = await this.getcolumnsbyschema(p)
            //console.log(r)
            r.result.rows.forEach(e => { schema[e.CNAME] = e })
            let cols = r.result.rows.map(e => { return `"${e.CNAME}"` })
            let vals = r.result.rows.map(e => {
                let r = `'{{e.CNAME}}'`
                if (e.CTYPE == 'JSON') { }
                return r
            })

            let r2 = await this.getdefaultcols(p)

            let defaultcols = r2.result.rows.map(e => e.cname)
            let r0 = await this.getpkcols(p)
            //console.log(r0)
            let pkcols = []
            if (r0.pkcols) { pkcols = r0.pkcols }
            let file = this.getPath(p, '/tables')
            let beforeInsert = 'beforeInsert:null'
            //let insertsql = `insert into "${schema}"."${tname}"(${cols.join(',')}) values()`
            try {
                let t = require(file + '.js')
                if (t.beforeInsert) {
                    beforeInsert = 'beforeInsert:' + t.beforeInsert.toString()
                }
            } catch (e) {
                // console.log(e)
            }
            let tmp = `
        module.exports = {
        cols: ${JSON.stringify(schema, null, "\t")},
        pkcols:${JSON.stringify(pkcols)},
        defaultcols:${JSON.stringify(defaultcols)},
        ${beforeInsert},
       /* beforeInsert:async function (db,data) {
            let r  = await db.getseq({seqname:'seqname'}) 
            //console.log(r)
            data.ID = r
            return data
        }
        */
    }
      `
            //console.log(file,tmp)
            fs.writeFileSync(file + '.js', tmp)
            utils.require2(file)
            return { code: 1, message: 'write ok' }
        } catch (e) {
            console.log(e)
            return { code: -3, message: e }
        }
    }

    async maketriggerdef(p) {
        let file = this.getPath(p, '/triggers')
        let exclude = []
        try {
            exclude = require(file).exclude
        } catch (e) {

        }
        if (!exclude) { exclude = [] }

        let r = await this.getcolumnsbyschema(p)
        let r0 = await this.getpkcols(p)
        //console.log(r)
        let tmp = `
         module.exports=
         {
             tname:'${p.tname}',
             uname:'${p.schema}',
             pkcols:${JSON.stringify(r0.pkcols)},
             exclude:${JSON.stringify(exclude)},
             cols: ${JSON.stringify(r.rows, null, '\t')}
         }
         `
        if (r.rows) {
            fs.writeFileSync(file + '.js', tmp)
            return { code: 1, message: '创建成功' }
        } else {
            return { code: -1, message: '无数据返回，创建失败' }
        }
    }

    filterCols(row, cols) {
        let row0 = {}
        cols.forEach(col => {
            if (row[col]) {
                row0[col] = row[col]
            }
        })
        return row0;
    }

    //按keys比较2个对象是否相等
    eqObjbykey({ o0, o1, keys }) {
        //console.log(arguments[0])
        return keys.every(key => o0[key] == o1[key])
    }

    //记录数据表操作日志
    async writelog(schema, tname, uname, otype, newdata, olddata) {

        let sql = `insert into "_zmh_dbop_log"(schema,tname,uname,otype,newdata,olddata,updata) values($1,$2,$3,$4,$5,$6,$7)`
        let updata = {}

        let t = await this.getschemafile({ schema, tname })
        //console.log(t)
        if (otype == 'UPDATE') {
            updata = olddata.map((o0) => {
                let o1 = newdata.filter(o1 => this.eqObjbykey({ o0, o1, keys: t.pkcols }))[0]
                let item = {}
                Object.keys(o0).filter(key => o0[key] != o1[key]).map(key => item[key] = [o0[key], o1[key]])
                let pk = {}
                t.pkcols.map(key => pk[key] = o0[key])
                return { pk, old: o0, updated: item }
            })
        }
        let r1 = await this.execute(sql, [schema, tname, uname, otype, JSON.stringify(newdata), JSON.stringify(olddata), JSON.stringify(updata)]);
        if (r1.code < 0) {
            let sql0 = `
            create SEQUENCE  "public"."_zmh_dbop_log_id_seq";
            cREATE TABLE "public"."_zmh_dbop_log" (
                            "id" int8 NOT NULL DEFAULT nextval('_zmh_dbop_log_id_seq'::regclass),
                            "schema" varchar(60) COLLATE "pg_catalog"."default",
                            "tname" varchar(60) COLLATE "pg_catalog"."default",
                            "uname" varchar(60) COLLATE "pg_catalog"."default",
                            "otype" varchar(10) COLLATE "pg_catalog"."default",
                            "olddata" json,
                            "newdata" json,
                            "opdate" timestamp(6) DEFAULT now(),
                            "updata" json,
                            CONSTRAINT "_zmh_dbop_log_pkey" PRIMARY KEY ("id")
             );
            `
            await this.execute(sql0);
            r1 = await this.execute(sql, [schema, tname, uname, otype, JSON.stringify(newdata), JSON.stringify(olddata), JSON.stringify(updata)]);
        }
        return r1
    }

    upperJSONKey(jsonObj) {
        let r0 = {}
        for (var key in jsonObj) {
            r0[key.toUpperCase()] = jsonObj[key];
        }
        return r0
    }
    lowerJSONKey(jsonObj) {
        let r0 = {}
        for (var key in jsonObj) {
            r0[key.toLowerCase()] = jsonObj[key];
        }
        return r0
    }



    async columnAdd({ tname, schema = 'public', cname, ctype = 'text', clen, cscale, cnull, cdef, ccomment }) {
        try {
            ctype = ctype.toLowerCase()
            if (ctype == 'varchar' || ctype == 'char') {
                ctype = ctype + '(' + clen + ')'
            }
            if (ctype == 'numeric') {
                ctype = ctype + '(' + clen + ',' + cscale + ')'
            }
            if (cnull == 'NO') {
                cnull = 'not null'
            } else {
                cnull = ''
            }
            if (cdef) {
                cdef = cdef.replace(/'/g, "''")
                cdef = `default '${cdef}'`
            } else {
                cdef = ''
            }
            let sql = `alter table "${schema}"."${tname}" add  "${cname}"  ${ctype} ${cnull} ${cdef}`
            let r = await this.execute(sql)
            if (ccomment) {
                ccomment = ccomment.replace(/'/g, "''")
                sql = `COMMENT ON COLUMN "${schema}"."${tname}"."${cname}" is  '${ccomment}'`
                r = await this.execute(sql)
            }
            return r
        } catch (e) {
            console.log(e)
            return { code: -1, err: e }
        }
    }
    async columnAddId({ tname, schema = 'public', cname = 'id', ctype = 'int8' }) {
        let sname = `${tname}_${cname}_seq`
        let sql = `
            CREATE SEQUENCE "${schema}"."${sname}" 
            INCREMENT 1
            MINVALUE  1
            MAXVALUE 9223372036854775807
            START 1
            CACHE 1;
       `
        let r = await this.execute(sql)
        sql = `alter table "${schema}"."${tname}" add  "${cname}"  ${ctype} default  nextval('${sname}'::regclass)`
        r = await this.execute(sql)
        return r
    }

    async columnDel({ tname, schema = 'public', cname }) {
        let sql = `alter table "${schema}"."${tname}" drop column "${cname}"`
        let r = await this.execute(sql)
        return r
    }

    async columnSetName({ tname, schema = 'public', cname, newname }) {
        let sql = `
           alter table "${schema}"."${tname}" rename column "${cname}"  to  "${newname}"
        `
        let r = await this.execute(sql)
        return r
    }
    async columnSet({ tname, schema = 'public', cname, ctype, clen, cscale, cnull, cdef, ccomment }) {
        let r
        let sql
        let r0 = await this.columnGet({ tname, schema, cname })
        if (r0.code == 1 && r0.data && r0.data.length > 0) {
            let d0 = r0.data[0]
            if (d0.cnull != cnull) {
                cnull = (cnull == 'NO') ? 'set' : 'drop'
                sql = `alter table "${schema}"."${tname}" alter "${cname}" ${cnull} not null `
                r = await this.execute(sql)
            }
            if (d0.ccomment != ccomment) {
                ccomment = ccomment.replace(/'/g, "''")
                ccomment = ccomment ? ccomment : ''
                sql = `COMMENT ON COLUMN "${schema}"."${tname}"."${cname}" is '${ccomment}'`
                r = await this.execute(sql)
            }
            if (d0.cdef != cdef) {
                if (!cdef) {
                    sql = `alter table "${schema}"."${tname}" alter "${cname}" drop default`
                } else {
                    cdef = cdef.replace(/'/g, "''")
                    sql = `alter table "${schema}"."${tname}" alter "${cname}" set default ${cdef} `
                }
                r = await this.execute(sql)
            }

            if (d0.ctype != ctype || d0.clen != clen || d0.cscale != cscale) {
                if (ctype == 'varchar' || ctype == 'char') {
                    ctype = ctype + '(' + clen + ')'
                }
                if (ctype == 'numeric') {
                    ctype = ctype + '(' + clen + ',' + cscale + ')'
                }
                sql = `alter table "${schema}"."${tname}" alter column "${cname}" type ${ctype} using("${cname}"::${ctype})`
                r = await this.execute(sql)
            }
        }
        return r
    }

    async columnGet({ tname, schema = 'public', cname }) {
        cname = (cname) ? `and a.attname ='${cname}'` : ''
        let sql = `
        select t1.cname cname, t1.descript ccomment ,t2.is_nullable cnull ,t1.data_type datatype,t2.udt_name ctype, 
            coalesce(t2.character_maximum_length,numeric_precision) clen,numeric_scale cscale, 
            t2.column_default cdef
        from (select a.attnum ,
            (select description from pg_catalog.pg_description where objoid=a.attrelid and objsubid=a.attnum) as descript ,
            a.attname cname,
            pg_catalog.format_type(a.atttypid,a.atttypmod) as data_type,
            b.sname,
            b.tname
           from 
             pg_catalog.pg_attribute a ,
             (select a1.oid,a1.relname tname,a2.nspname sname from pg_class a1, pg_namespace a2  where a1.relname='${tname}' and a1.relnamespace = a2.oid and 
                  a2.nspname = '${schema}') as b
            where a.attrelid=b.oid and a.attnum>=0 and not a.attisdropped ${cname}
            ) t1,
            information_schema.columns t2
              where t1.sname =t2.table_schema and t1.tname = t2.table_name and t2.column_name = t1.cname
        `
        let r = await this.execute(sql)
        let r0 = await this.PKGet0({ tname, schema })

        r = (r.code == 1) ? { code: r.code, data: r.result.rows } : r
        if (r.code == 1) {
            r.data = r.data.map(i => {
                switch (i.ctype.substring(0, 3)) {
                    case 'int':
                    case 'num':
                        i.jstype = 'number';
                        break;
                    case 'json':
                        i.jstype = 'object';
                        break;
                    case 'dat':
                    case 'tim':
                        i.jstype = 'date';
                        break;
                    case 'boo':
                        i.jstype = 'boolean';
                        break;
                    default:
                        i.jstype = 'string';
                }
                (r0.data.filter(pk => i.cname == pk.colname).length > 0) ? i.pk = true : i.pk = false;

                return i
            })
        }
        return r
    }


    async indexAdd({ tname, schema = 'public', idxName, cols = [], op = '', unique = '' }) {
        if (op != '') {
            op = ' USING ' + op + ' ';
        }
        let sql = `
       create ${unique} index  "${idxName}"  on "${schema}"."${tname}"${op}(${'"' + cols.join('","') + '"'} )
       `
        console.log(sql)
        //CREATE UNIQUE INDEX "test_table_pid_index" ON "public"."test_table" USING btree ("pid");
        let r = await this.execute(sql)
        console.log(r)
        return r
    }

    async indexDel({ schema = 'public', idxName }) {
        let sql = `
        drop index "${schema}"."${idxName}"
        `
        let r = await this.execute(sql)
        //console.log(r)      
        return r
    }

    async indexGet({ tname, schema = 'public' }) {
        let sql = `
        select
        A.SCHEMANAME,
        A.TABLENAME,
        A.INDEXNAME,
        A.TABLESPACE,
        A.INDEXDEF,
        B.AMNAME,
        C.INDEXRELID,
        C.INDNATTS,
        C.INDISUNIQUE,
        C.INDISPRIMARY,
        C.INDISCLUSTERED,
        D.DESCRIPTION
        from
            PG_AM B 
            left join PG_CLASS F on
                B.OID = F.RELAM 
            left join PG_STAT_ALL_INDEXES E on
                F.OID = E.INDEXRELID 
            left join PG_INDEX C on
                E.INDEXRELID = C.INDEXRELID 
            left outer join PG_DESCRIPTION D on
                C.INDEXRELID = D.OBJOID,
            PG_INDEXES A
        where
        A.SCHEMANAME = E.SCHEMANAME
        and A.TABLENAME = E.RELNAME
        and A.INDEXNAME = E.INDEXRELNAME
        and E.SCHEMANAME = '${schema}'
        and E.RELNAME = '${tname}'
        and c.indisprimary =false        
        `
        let r = await this.execute(sql)
        r = (r.code == 1) ? { code: r.code, data: r.result.rows } : r
        return r
    }



    async PKAdd({ tname, schema = 'public', cols = ['id'] }) {

        this.PKDel({ tname, schema })
        let sql = `ALTER TABLE "${schema}"."${tname}" add primary key (${cols.join(',')})`
        let r = await this.execute(sql)
        //console.log(r)      
        r = (r.code == 1) ? { code: 1, data: r.result.rows } : { code: -1, error: 'data err' }
        return r
    }



    async PKGet({ tname, schema = 'public' }) {
        let sql = `
        select sname,tname,pkname ,array_to_string(array_agg(colname),',') cols from 
        (select pg_constraint.conname pkname,pg_attribute.attname as colname ,t.nspname sname,t.relname tname
  from pg_constraint  
        inner join (select a1.oid,a1.relname,a2.nspname from pg_class a1,pg_namespace a2 
               where a1.relnamespace =a2.oid and a1.relname = '${tname}' and a2.nspname ='${schema}' ) t
      on pg_constraint.conrelid = t.oid 
    inner join pg_attribute 
            on pg_attribute.attrelid = t.oid and pg_constraint.conkey @>array[pg_attribute.attnum ]
    inner join pg_type on pg_type.oid = pg_attribute.atttypid
  where pg_constraint.contype='p')  t     
        group by sname,tname,pkname        
        `
        let r = await this.execute(sql);
        r = (r.code == 1 && r.result.rowCount > 0) ? { code: 1, data: r.result.rows } : { code: -1, error: 'no data' }
        return r
        //{code:1,data:r.result.rows}
    }

    async PKGet0({ tname, schema = 'public' }) {
        let sql = `
        select sname,tname,pkname ,colname from 
        (select pg_constraint.conname pkname,pg_attribute.attname as colname ,t.nspname sname,t.relname tname
  from pg_constraint  
        inner join (select a1.oid,a1.relname,a2.nspname from pg_class a1,pg_namespace a2 
               where a1.relnamespace =a2.oid and a1.relname = '${tname}' and a2.nspname ='${schema}' ) t
      on pg_constraint.conrelid = t.oid 
    inner join pg_attribute 
            on pg_attribute.attrelid = t.oid and pg_constraint.conkey @>array[pg_attribute.attnum ]
    inner join pg_type on pg_type.oid = pg_attribute.atttypid
  where pg_constraint.contype='p')  t     
        `
        let r = await this.execute(sql);
        r = (r.code == 1 && r.result.rowCount > 0) ? { code: 1, data: r.result.rows } : { code: -1, error: 'no data' }
        return r
        //{code:1,data:r.result.rows}
    }


    async PKDel({ tname, schema = 'public' }) {
        let r = await this.PKGet({ tname, schema });
        console.log(r)
        if (r.code == 1 && r.data) {
            let sql = `alter table "${schema}"."${tname}" drop CONSTRAINT "${r.data[0].pkname}"`
            console.log(sql)
            r = await this.execute(sql)
            console.log(r)
            r = (r.code == 1) ? { code: 1 } : r
            return r
        } else {
            return { code: -1, error: 'PK del error' }
        }
    }


    async tableGet({ schema }) {

        schema = schema ? `where table_schema='${schema}'` : ''
        let sql = `select * from information_schema.tables ${schema}`
        console.log(sql)
        let r = await this.execute(sql)
        //console.log(r.result.rows)      
        r = (r.code == 1) ? { code: r.code, data: r.result.rows } : r
        return r
    }


    async tableCopy({ tname, newname, schema = 'public', data }) {
        let sql
        let r
        sql = `drop table if exists "${schema}"."${newname}"`
        await this.execute(sql);
        sql = `create  table  "${schema}"."${newname}"  (LIKE  "${schema}"."${tname}" INCLUDING ALL)`
        r = await this.execute(sql);
        if (data == 1) {
            sql = `INSERT INTO "${schema}"."${newname}" select * from "${schema}"."${tname}"`
            r = await this.execute(sql);
        }
        return r
    }

    async tableCreate({ tname, schema = 'public' }) {
        let sql
        let r
        sql = `create SEQUENCE "${schema}"."${tname}_id_seq" `
        r = await this.execute(sql);
        sql = `create  table  "${schema}"."${tname}"  (
            "id" int8 NOT NULL DEFAULT nextval('${schema}.${tname}_id_seq'::regclass),
            CONSTRAINT "${tname}_pkey" PRIMARY KEY ("id")
        )`
        r = await this.execute(sql);
        return r

    }

    async tableCreateSys({ tname, schema = 'public' }) {
        let sql
        let r
        sql = `create SEQUENCE "${schema}"."${tname}_id_seq" `
        r = await this.execute(sql);
        sql = `create  table  "${schema}"."${tname}"  (
            "id" int8 NOT NULL DEFAULT nextval('${tname}_id_seq'::regclass),
            CONSTRAINT "${tname}_pkey" PRIMARY KEY ("id")
        )`
        r = await this.execute(sql);
        return r

    }



    async tableRename({ tname, newname, schema = 'public' }) {

        let sql
        sql = `alter table "${schema}"."${tname}" rename to "${newname}"`
        let r = await this.execute(sql);
        return r

    }

    async tableDel({ tname, schema = 'public' }) {
        let sql
        sql = `drop table if exists "${schema}"."${tname}" ;`
        let r = await this.execute(sql);
        return r
    }





    async formInsert(p) {
        if (!p.data) { return { code: 1, data: [] } }
        if (JSON.stringify(p.data) == '{}' || JSON.stringify(p.data) == '[]') { return { code: 1, data: [] } }
        //对象转数组
        if (!utils.isArray(p.data)) {
            p.data = [p.data]
        }
        //转换JSON KEY的大小写
        if (p.case == 'upper') {
            p.data = utils.upperJSONKey(p.data)
        }
        if (p.case == 'lower') {
            p.data = utils.lowerJSONKey(p.data)
        }
        let sql = `insert into sys_formdata(tname,data) select '${p.tname}',"value" from jsonb_array_elements('${JSON.stringify(p.data)}') returning *`
        let r1 = await this.execute(sql)
        if (r1.code == 1) {
            let data0 = r1.result.rows.map(i => { return { ...i.data, __id: i.__id } })
            return { code: 1, data: data0 }
        } else {
            return r1
        }
    }

    async formDelete(p) {
        if (!p.data) { return { code: 1, data: [] } }
        if (JSON.stringify(p.data) == '{}' || JSON.stringify(p.data) == '[]') { return { code: 1, data: [] } }
        //对象转数组
        if (!utils.isArray(p.data)) {
            p.data = [p.data]
        }
        let sql = `delete  from sys_formdata a using  (select ("value"->>'__id')::bigint as "id"  from jsonb_array_elements('${JSON.stringify(p.data)}')) b  where __id =b.id`
        let r1 = await this.execute(sql)
        //console.log(r1)
        if (r1.code == 1) {
            if (r1.result.rowCount > 0) {
                return { code: 1, data: p.data }
            } else {
                return { code: 1, data: null }
            }
        } else {
            return r1
        }
    }


    async formGet(p) {

        let sql = `select * from sys_formdata where tname ='${p.tname}'  and __id =${p.__id} `
        let r1 = await this.execute(sql)
        if (r1.code == 1) {
            let data0 = r1.result.rows.map(i => { return { ...i.data, __id: i.__id } })
            return { code: 1, data: data0 }
        } else {
            return r1
        }

    }

    async formUpdate(p) {
        if (!p.data) { return { code: 1, data: [] } }
        if (JSON.stringify(p.data) == '{}' || JSON.stringify(p.data) == '[]') { return { code: 1, data: [] } }
        //对象转数组
        if (!utils.isArray(p.data)) {
            p.data = [p.data]
        }
        //转换JSON KEY的大小写
        if (p.case == 'upper') {
            p.data = utils.upperJSONKey(p.data)
        }
        if (p.case == 'lower') {
            p.data = utils.lowerJSONKey(p.data)
        }
        let sql = `update sys_formdata a set data = a.data||b.data::jsonb from  
                    (select "value" as "data" from jsonb_array_elements('${JSON.stringify(p.data)}')) b
                    where  a.tname ='${p.tname}' and a.__id = (b."data"->>'__id')::bigint
                  returning a.*`
        let r1 = await this.execute(sql)
        if (r1.code == 1) {
            let data0 = r1.result.rows.map(i => { return { ...i.data, __id: i.__id } })
            return { code: 1, data: data0 }
        } else {
            return r1
        }
    }

    async formSelect(p) {

        let sqla = `select __id,b.*  from sys_formdata a join lateral (select * from  jsonb_populate_record(null::form_${p.tname},a.data) ) b  on tname ='${p.tname}'`
        let filters = ``
        let sorts = ``
        let fields = '*'
        let pages = ''
        //设置字段列表
        if (p.fields) { fields = p.fields.map(i => `"${i}"`).join(',') }
        //设置条件过滤表达式
        if (p.filter && p.filter.filters) { filters = ' where ' + this.getwhere(p.filter) }
        //设置排序
        if (p.sort) { sorts = ' order by ' + p.sort.map(i => `"${i.field}" ${i.dir}`).join(',') }
        //设置分页
        if (p.pagesize) { p.pageSize = p.pagesize }
        if (!p.pageSize) { p.pageSize = 10 }
        if (!p.page) { p.page = 1 }
        pages = `limit ${p.pageSize} offset ${(p.page - 1) * p.pageSize}`
        //运行数据查询
        let sql = `select ${fields} from (${sqla}) as _tt0 ${filters} ${sorts} ${pages} `
        let r0 = await this.execute(sql)
        //统计符合记录总数
        let sql1 = `select count(1) as total from (${sqla}) as _tt0 ${filters}`
        let r1 = await this.execute(sql1)
        let r = {}
        if (r1.code == 1 && r0.code == 1) {
            r = { code: 1, data: r0.result.rows, total: r1.result.rows[0].total }
        }
        return r
    }


    async tableInsert({ schema = 'public', tname, data, keycase, user, log }) {
        let p = arguments[0]
        //console.log(p)
        try {
            //转换JSON KEY的大小写
            //console.log(p)
            if (!data) { return { code: 1, data: [] } }
            if (JSON.stringify(data) == '{}' || JSON.stringify(data) == '[]') { return { code: 1, data: [] } }

            if (keycase == 'upper') { data = utils.upperJSONKey(data) }
            if (keycase == 'lower') { data = utils.lowerJSONKey(data) }
            //对象转数组
            if (!utils.isArray(data)) { data = [data] }
            //根据类型转换数据格式
            data = (await this.dtproc(p)).data
            //获取表定义文件
            let t = await this.getschemafile(p)
            //console.log(t)
            //构造主键清单
            //let cs = t.pkcols.map(cname => `"${cname}"`).join(',')
            //设置默认值空字段 为default
            //构造字段清单
            let cs0 = Object.keys(t.cols).map(cname => `"${cname}"`).join(',')
            //let cs1 = Object.keys(data[0]).map(cname => `EXCLUDED."${cname}"`).join(',')
            //构造values
            let val = data.map(row => {
                let vals = []
                for (let col in t.cols) {
                    //是否存在col属性
                    if (Reflect.has(row, col)) {
                        //格式化变量值
                        if (row[col] == null) { vals.push('default') } else {
                            if (typeof (row[col]) == 'string') {
                                vals.push(`'${row[col].replace(/'/g, "''")}'`)
                            } else {
                                if (typeof (row[col]) == 'object') {
                                    //let rr =`'${JSON.stringify(row[col])}'`
                                    vals.push("'"+`${JSON.stringify(row[col])}`.replace(/'/g, "''")+"'")
                                } else {
                                    vals.push(`'${row[col]}'`)
                                }
                            }
                        }
                    } else {
                        vals.push('default')
                    }
                }
                return '(' + vals.join(',') + ')'
            })

            let sql = `insert into "${schema}"."${tname}"(${cs0}) values${val.join(',')}  returning *`
            console.log(sql)
            let r1 = await this.execute(sql)
            let uname = 'admin'
            if (user && user.name) { uname = user.name }
            if (log && r1.code == 1 && r1.result.rowCount > 0) {
                await this.writelog(schema, tname, uname, 'INSERT', r1.result.rows, null)
            }
            if (r1.code == 1 && r1.result.rowCount > 0) { return { code: 1, data: r1.result.rows } }
            return r1
        } catch (err) {
            console.log(err)
            return { code: -3, err }
        }
    }




    async tableUpdate({ schema = 'public', tname, data, keycase, user, log }) {
        let p = arguments[0]
        try {

            if (!data) { return { code: 1, data: [] } }
            if (JSON.stringify(data) == '{}' || JSON.stringify(data) == '[]') { return { code: 1, data: [] } }
            //对象转数组
            if (!utils.isArray(data)) { data = [data] }
            //转换JSON KEY的大小写
            if (keycase == 'upper') { data = utils.upperJSONKey(data) }
            if (keycase == 'lower') { data = utils.lowerJSONKey(data) }
            //转换时间格式
            data = (await this.dtproc(p)).data
            let t = await this.getschemafile(p)
            //tname =`"${schema}"."${tname}"`
            //构造where字段
            let wheres = t.pkcols.map(cname => `a."${cname}" = b."${cname}"`).join(' and ')
            //构造set字段清单
            let sets = Object.keys(data[0]).filter(key => (t.pkcols.indexOf(key) < 0)).filter(key => t.cols[key]).map(cname => `"${cname}" = b."${cname}"`).join(',')

            let uname = 'admin'
            if (user && user.name) { uname = user.name }

            let sql = `update  "${schema}"."${tname}" a set ${sets}
                 from (select * from json_populate_recordset(null::"${schema}"."${tname}",'${JSON.stringify(data).replace(/'/g, "''")}')) b where ${wheres} returning a.*`
            //console.log(sql) 
            let olddata = []
            if (log) {
                //console.log('111',p)
                let rr = await this.getTableData(p)
                if (rr.code == 1) { olddata = rr.data }
                //console.log(rr)
            }
            //console.log(sql)
            let r1 = await this.execute(sql, [])
            //console.log(r1)
            if (log && r1.code == 1 && r1.result.rowCount > 0) {
                await this.writelog(schema, tname, uname, 'UPDATE', r1.result.rows, olddata)
            }
            if (r1.code == 1 && r1.result.rowCount > 0) { return { code: 1, data: r1.result.rows } }
            return r1
        } catch (err) {
            console.log(err)
            return { code: -3, err }

        }
    }

    async tableUpsert({ schema = 'public', tname, data, keycase, user, log }) {
        let p = arguments[0]
        try {
            //转换JSON KEY的大小写
            if (!data) { return { code: 1, data: [] } }
            if (JSON.stringify(data) == '{}' || JSON.stringify(data) == '[]') { return { code: 1, data: [] } }

            //对象转数组
            if (!utils.isArray(data)) { data = [data] }
            //转换JSON KEY的大小写
            if (keycase == 'upper') { data = utils.upperJSONKey(data) }
            if (keycase == 'lower') { data = utils.lowerJSONKey(data) }
            //转换时间格式
            data = (await this.dtproc(p)).data
            let t = await this.getschemafile(p)
            //构造主键清单
            let cs = t.pkcols.map(cname => `"${cname}"`).join(',')

            //构造字段清单
            let cnames =Object.keys(data[0]).filter(i=>t.cols[i])
            let cs0 = Object.keys(t.cols).map(cname => `"${cname}"`).join(',')
            let cs00 = cnames.map(cname => `"${cname}"`).join(',')
            
            let cs1 = cnames.map(cname => `EXCLUDED."${cname}"`).join(',')

            //构造values
            let val = data.map(row => {
                let vals = []
                for (let col in t.cols) {
                    //是否存在col属性
                    if (Reflect.has(row, col)) {
                        //格式化变量值
                        if (row[col] == null) { vals.push('default') } else {
                            if (typeof (row[col]) == 'string') {
                                vals.push(`'${row[col].replace(/'/g, "''")}'`)
                            } else {
                                if (typeof (row[col]) == 'object') {
                                    vals.push(`'${JSON.stringify(row[col])}'`)
                                } else {
                                    vals.push(`'${row[col]}'`)
                                }
                            }
                        }
                    } else {
                        vals.push('default')
                    }
                }
                return '(' + vals.join(',') + ')'
            })

            let sql = `insert into "${schema}"."${tname}"(${cs0}) values${val.join(',')}  ON CONFLICT (${cs}) DO update set (${cs00}) = (${cs1}) returning *`

            //console.log(sql)
            let r1 = await this.execute(sql)
            let uname = 'admin'
            if (user && user.name) { uname = user.name }
            if (log && r1.code == 1 && r1.result.rowCount) {
                await this.writelog(p.schema, p.tname, uname, 'UPSERT', r1.rows, data)
            }
            if (r1.code == 1) { return { code: 1, data: r1.result.rows } }
            return r1
        } catch (e) {
            console.log(e)
            return { code: -3, err }

        }
    }


    async tableDelete({ schema = 'public', tname, data, keycase, user, log }) {

        //转换JSON KEY的大小写
        //转换JSON KEY的大小写
        try {
            if (!data) { return { code: 1, data: [] } }
            if (JSON.stringify(data) == '{}' || JSON.stringify(data) == '[]') { return { code: 1, data: [] } }
            if (keycase == 'upper') {
                data = utils.upperJSONKey(data)
            }
            if (keycase == 'lower') {
                data = utils.lowerJSONKey(data)
            }
            //对象转数组
            if (!utils.isArray(data)) {
                data = [data]
            }
            //转换时间格式
            data = await this.dtproc(data)
            let t = await this.getschemafile({ schema, tname })
            let wheres = t.pkcols.map(cname => `a."${cname}"=b."${cname}"`).join(' and ')
            let uname = 'admin'
            if (user && user.name) { uname = user.name }
            let sql = `delete from  "${schema}"."${tname}" a using (select * from json_populate_recordset(null::"${schema}"."${tname}",$1)) b 
               where ${wheres} returning *`
            //console.log(sql)
            //console.log(data)

            let r1 = await this.execute(sql, [JSON.stringify(data)])

            if (log && r1.code && r1.result.rowCount) {
                await this.writelog(schema, tname, uname, 'DELETE', null, r1.result.rows)
            }
            if (r1.code == 1) { return { code: 1, data: r1.result.rows } }
            return r1
        } catch (err) {
            console.log(err)
            return ({ code: -3, err })
        }
    }


    async getTableData({ schema = 'public', tname, data }) {
        let p = arguments[0]
        console.log(p)
        let t = await this.getschemafile(p)
        tname = `"${schema}"."${tname}"`

        let wheres = t.pkcols.map(cname => `a."${cname}"=b."${cname}"`).join(' and ')
        let sql = `select  a.* from  ${tname} a , (select * from json_populate_recordset(null::${tname},'${JSON.stringify(data)}'::json)) b 
           where ${wheres}`
        let r1 = await this.execute(sql)
        if (r1.code == 1) { return { code: 1, data: r1.result.rows } }
        return r1
    }

    async tableSelect({ schema = 'public', tname, page = 1, pageSize = 20000, sort, filter, fields }) {
        let p = arguments[0]
        try {

            //console.log(p)
            //设置字段列表
            let cs = '*'
            if (fields) { field0 = fields.map(i => `"${i}"`).join(',') }
            //设置条件过滤表达式
            let filters = ``
            if (filter && filter.filters) { filters = ' where ' + this.getwhere(filter) }
            //设置排序
            let sorts = ``
            if (sort && sort.length > 0) { sorts = ' order by ' + sort.map(i => `"${i.field}" ${i.dir}`).join(',') }
            //设置分页
            let pages = ''
            pages = `limit ${pageSize} offset ${(page - 1) * pageSize}`
            //运行数据查询
            let sql = `select ${cs} from "${schema}"."${tname}" ${filters} ${sorts} ${pages} `
            //console.log(sql)
            let r0 = await this.execute(sql)
            //console.log(r0)
            //统计符合记录总数
            let sql1 = `select count(1) as total from "${schema}"."${tname}" ${filters}`
            let r1 = await this.execute(sql1)
            let r = {}
            if (r1.code == 1 && r0.code == 1) {
                return { code: 1, data: r0.result.rows, total: r1.result.rows[0].total }
            } else {
                return { code: -3, err: '数据错误' }
            }
        } catch (err) {
            console.log(err)
            return { code: -3, err }
        }
    }

    async sqlSelect(p = {}) {
        try {

            let sqla = await this.getsqlfile(p)
            //console.log(sqla)
            //console.log(sqla)
            if (sqla) { sqla = sqla.render(p.param) }
            let filters = ``
            let sorts = ``
            let fields = '*'
            let pages = ''
            //设置字段列表
            if (p.fields) { fields = p.fields.map(i => `"${i}"`).join(',') }
            //设置条件过滤表达式
            if (p.filter && p.filter.filters) { filters = ' where ' + this.getwhere(p.filter) }
            //设置排序
            if (p.sort) { sorts = ' order by ' + p.sort.map(i => `"${i.field}" ${i.dir}`).join(',') }
            //设置分页
            if (p.pagesize) { p.pageSize = p.pagesize }
            if (!p.pageSize) { p.pageSize = 10 }
            if (!p.page) { p.page = 1 }
            pages = `limit ${p.pageSize} offset ${(p.page - 1) * p.pageSize}`
            //运行数据查询
            let sql = `select ${fields} from (${sqla}) as _tt0 ${filters} ${sorts} ${pages} `
            let r0 = await this.execute(sql)
            //统计符合记录总数
            let sql1 = `select count(1) as total from (${sqla}) as _tt0 ${filters}`
            let r1 = await this.execute(sql1)
            let r = {}
            if (r1.code == 1 && r0.code == 1) {
                r = { code: 1, data: r0.result.rows, total: r1.result.rows[0].total }
            }
            return r
        } catch (err) {
            console.log(err)
            return { code: -3, err }
        }
    }

}

module.exports = Db;
